#!/usr/bin/perl 
#
# dnsblcheck.pl 
# 02.27.2006 Paul Venezia
#
# Large parts of the milter interface code by Evan Harris as part of relaydelay.pl
#
#
#
#

use Sendmail::Milter;
use Socket;
use POSIX qw(strftime);
use Errno qw(ENOENT);
use Net::Nslookup;

use DBI;
use FileHandle;

use strict;

# If you do/don't want to see debugging messages in the logs,
#   then set this appropriately.
my $verbose = 1;
my $run_as_daemon = 1; # Default is to run as a daemon
my $log_file = '/var/log/dnsblcheck.log';    

# Database connection params
my $database_type = 'mysql';
my $database_name = 'dnsblcheck';
my $database_host = 'localhost';
my $database_port = 3306;
my $database_user = 'db_user';
my $database_pass = 'db_pass';

# Set this to indicate the milter "name" that this milter will be 
#   identified by.  This must match the first parameter from the 
#   INPUT_MAIL_FILTER definition in the sendmail.mc configuration.
my $milter_filter_name = 'dnsblcheck';

# This parameter determines how the milter interfaces with the libmilter
#   API.  Normally, if using a milter on the same machine that is running
#   sendmail, it will be something like 'local:/var/run/dnsblcheck.sock', 
#   but if you want to run the milter on a different machine than is running 
#   sendmail, you will need to specify how to connect to that copy of 
#   sendmail by setting this to indicate the machine and port that the 
#   remote sendmail is listening for connections on with something 
#   similar to 'inet:2526@sendmail.server.org'.
# This parameter must match the S= option in the INPUT_MAIL_FILTER
#   definition in the sendmail.mc file.
my $milter_socket_connection = 'local:/var/spool/MIMEDefang/dnsblcheck.sock';

# Where the pid file should be stored for dnsblcheck
my $dnsblcheck_pid_file = '/var/run/dnsblcheck.pid';

# Set this to something nonzero to limit the number of children that the 
#   milter will spawn.  Since children are never recycled (there seems 
#   to be a problem doing that with Sendmail::Milter), threads,
#   once created, will exist until the milter is shutdown.  Each thread
#   also consumes a database connection, so limiting db connections and
#   memory footprint are both good reasons to set this.
# If your mail server handles a large amount of mail, you may need to 
#   increase this limit to avoid blocking, but the default limit is
#   already pretty high, and should be sufficient for all but very 
#   large sites.
# Setting to zero makes the number of threads unlimited.
my $maximum_milter_threads = 40;

#############################################################
# End of options for use in external config file
#############################################################


# Global vars that should probably not be in the external config file
my $global_dbh;
  
  if ($log_file) { # Keep the output in a log file
    open STDOUT, ">>$log_file" or die "Couldn't redirect STDOUT to $log_file: $!";
    STDOUT->autoflush(1);
  }

#######################################################################
# Database functions
#######################################################################

sub db_connect($) {
  my $verbose = shift;

  return $global_dbh if (defined $global_dbh);

  my $dsn = "DBI:$database_type:database=$database_name:host=$database_host:port=$database_port";
  print "DBI Connecting to $dsn\n" if $verbose;

  # Note: We do all manual error checking for db errors
  my $dbh = DBI->connect($dsn, $database_user, $database_pass, 
                         { PrintError => 0, RaiseError => $verbose });

  $global_dbh = $dbh;
  return $global_dbh;
}

sub db_disconnect {
  $global_dbh->disconnect() if (defined $global_dbh);
  $global_dbh = undef;
  return 0;
}


#############################################################################
#
# Milter Callback Functions:
#
#  Each of these callbacks is actually called with a first argument
#  that is blessed into the pseudo-package Sendmail::Milter::Context. You can
#  use them like object methods of package Sendmail::Milter::Context.
#
#  $ctx is a blessed reference of package Sendmail::Milter::Context to something
#  yucky, but the Mail Filter API routines are available as object methods
#  (sans the smfi_ prefix) from this
#
#############################################################################

# Here we perform the bulk of the work, since here we have individual recipient
#   information, and can act on it.

sub envrcpt_callback
{
  my $ctx = shift;
  my @args = @_;

  # Get the time in seconds
  my $timestamp = time();

  # Get the hostname (needs a module that is not necessarily installed)
  #   Not used (since I don't want to depend on it)
  #my $hostname = hostname();

  print strftime("\n=== %Y-%m-%d %H:%M:%S ===\n", localtime($timestamp)) if ($verbose);

  # Get the database handle (after got the privdata)
  my $dbh = db_connect(0) or goto DB_FAILURE;
  
  # Get the remote hostname and ip in the form "[ident@][hostname] [ip]"
  my $tmp = $ctx->getsymval("{_}");  
  my ($relay_ip, $relay_name, $relay_ident, $relay_maybe_forged);
  if ($tmp =~ /\A(\S*@|)(\S*) ?\[(.*)\]( \(may be forged\)|)\Z/) {
    $relay_ident = $1;
    $relay_name = $2;
    $relay_ip = $3;
    $relay_maybe_forged = (length($4) > 0 ? 1 : 0);
  }
  my $relay_name_reversed = reverse($relay_name);
        
  # Collect the rest of the info for our checks
  my $mail_mailer = $ctx->getsymval("{mail_mailer}");
  my $sender      = $ctx->getsymval("{mail_addr}");
  my $rcpt_mailer = $ctx->getsymval("{rcpt_mailer}");
  my $recipient   = $ctx->getsymval("{rcpt_addr}");
  my $queue_id    = $ctx->getsymval("{i}");
  my $authen      = $ctx->getsymval("{auth_authen}");
  my $authtype    = $ctx->getsymval("{auth_type}");
  my $ifaddr      = $ctx->getsymval("{if_addr}");

  print "  Relay: $tmp - If_Addr: $ifaddr\n" if ($verbose);
  print "  RelayIP: $relay_ip - RelayName: $relay_name - RelayIdent: $relay_ident - PossiblyForged: $relay_maybe_forged\n" if ($verbose);
  print "  From: $sender - To: $recipient\n" if ($verbose);
  print "  InMailer: $mail_mailer - OutMailer: $rcpt_mailer - QueueID: $queue_id\n" if ($verbose);

  # Only do our processing if the inbound mailer is an smtp variant.
  #   A lot of spam is sent with the null sender address <>.  Sendmail reports 
  #   that and other "local looking" from addresses as using the local mailer, 
  #   even though they are coming from off-site.  So we have to exclude the 
  #   "local" mailer from the checks since it lies.
  if (($mail_mailer !~ /smtp\Z/i) and ($mail_mailer !~ /\Alocal\Z/i)) {
    # we aren't using an smtp-like mailer, so bypass checks
    print "  Mail delivery is not using an smtp-like mailer.  Skipping checks.\n" if ($verbose);
    return SMFIS_CONTINUE;
  }

  # Check to see if the mail is looped back on a local interface and skip checks if so
  if ($ifaddr eq $relay_ip) {
    print "  Mail delivery is sent from a local interface.  Skipping checks.\n" if ($verbose);
    return SMFIS_CONTINUE;
  }

  # Only do our processing if the mail client is not authenticated in some way
  if (defined($authen) and $authen ne "")
  {
    print "  AuthType: $authtype - Credentials: $authen\n" if ($verbose);
    print "  Mail delivery is authenticated.  Skipping checks.\n" if ($verbose);
    return SMFIS_CONTINUE;
  }


my @bl=("bl.spamcop.net","dnsbl.sorbs.net","sbl-xbl.spamhaus.org");

my $totquery = "UPDATE dnsbls SET total=total+1 WHERE dnsbl = 'total'";
            my $sth = $dbh->prepare($totquery) or goto DB_FAILURE;
            $sth->execute() or goto DB_FAILURE;
            goto DB_FAILURE if ($sth->err);
            $sth->finish();

foreach my $thisbl (<@bl>) {
        chomp($relay_ip);
        my $revip = $relay_ip;
        $revip =~ s/(\d+)\.(\d+)\.(\d+)\.(\d+)/$4\.$3\.$2\.$1/;
        my $host = "$revip.$thisbl";
        my $status = nslookup(host=>"$host", type=> "A");
        if ($status) {
        	print "$thisbl - $status\n" if $verbose;

            my $blquery = "UPDATE dnsbls SET `$status`=`$status` + 1, total=total+1 WHERE dnsbl = '$thisbl'";
            my $sth = $dbh->prepare($blquery);
            $sth->execute();
            $sth->finish();
        }

}

db_disconnect();
print "Passing on..." if $verbose;
return SMFIS_CONTINUE;

}

sub catch_sig {
  my $signame = shift;
  print "Got a SIG$signame.\nClosing DB connection.\n" if $verbose;
  db_disconnect();
  print "Exiting dnsblcheck daemon process.\n";
  close(STDOUT);
  exit 0;
}

my %my_callbacks =
(
#	'connect' =>	\&connect_callback,
#	'helo' =>	\&helo_callback,
#	'envfrom' =>	\&envfrom_callback,
	'envrcpt' =>	\&envrcpt_callback,
#	'header' =>	\&header_callback,
#	'eoh' =>	\&eoh_callback,
#	'body' =>	\&body_callback,
#	'eom' =>	\&eom_callback,
#	'abort' =>	\&abort_callback,
#	'close' =>	\&close_callback,
);

BEGIN:
{

  # If the config file was specified on the command line, use it

  print "Using connection '$milter_socket_connection' for filter $milter_filter_name\n" if $verbose;

  if ($milter_socket_connection =~ /^local:(.+)$/i) {
    my $unix_socket = $1;

    if (-e $unix_socket) {
      print "Attempting to unlink local UNIX socket '$unix_socket' ... " if $verbose;

      if (unlink($unix_socket) == 0) {
        print "failed.\n" if $verbose;
        exit;
      }
      print "successful.\n" if $verbose;
    }
  }

  if (not Sendmail::Milter::setconn("$milter_socket_connection")) {
    die "Failed to set up connection: $?\n";
#    exit;
  }

  # Make sure we can connect to the database 
  my $dbh = db_connect(1);
  die "$DBI::errstr\n" unless($dbh);
  # and disconnect again, since the callbacks won't have access to the handle
  db_disconnect();

  #
  #  The flags parameter is optional. SMFI_CURR_ACTS sets all of the
  #  current version's filtering capabilities.
  #

  if (not Sendmail::Milter::register("$milter_filter_name", \%my_callbacks, SMFI_CURR_ACTS)) {
    die "Failed to register callbacks for $milter_filter_name.\n";
#    exit;
  }

  if ($run_as_daemon) {
    if (not $log_file) {print "Warning: Running as a daemon, but output has not been redirected to a log file.\n";}


# I don't think we have to worry about reaping zombies since we're only 
# spawning a single child process then exiting:
#    $SIG{CHLD} = 'IGNORE'; # Automatically reap children
    defined(my $child_pid = fork()) or die "Couldn't fork daemon process:$!";
    if ($child_pid) {
      # I must be the parent:
      print "Spawned dnsblcheck daemon process $child_pid.\n" if $verbose;
      exit 0;
    }
  }

  # I must be the child (or using foreground operation):
  # Record pid to file
  if (defined $dnsblcheck_pid_file) {
    open(PIDF, ">$dnsblcheck_pid_file") ||
      die "Unable to record PID to '$dnsblcheck_pid_file': $!\n";
    print PIDF "$$\n";
    close PIDF;
  }

  if ($run_as_daemon) {
    # Be a nice daemon:
    POSIX::setsid or die "Couldn't start a new session: $!";
    chdir '/' or die "Couldn't chdir to /: $!";
    open STDIN, '/dev/null' or die "Couldn't redirect STDIN from /dev/null: $!";
    open STDERR, '>&STDOUT' or die "Couldn't dup STDOUT: $!";
    my $sigset = POSIX::SigSet->new();
    my $action = POSIX::SigAction->new('catch_sig',$sigset,&POSIX::SA_NODEFER);
    POSIX::sigaction(&POSIX::SIGQUIT, $action);
  }

  print "Starting Sendmail::Milter $Sendmail::Milter::VERSION engine.\n";

  # Parameters to main are max num of interpreters, num requests to service before recycling threads
  # We don't set it to recycle children, as that seems to cause coredumps.
  if (Sendmail::Milter::main($maximum_milter_threads, 0)) {
    print "Successful exit from the Sendmail::Milter engine.\n";
  }
  else {
    print "Unsuccessful exit from the Sendmail::Milter engine.\n";
  }
}


# Make sure when threads are recycled that we release the global db connection
END {
  if (not $run_as_daemon) { # Signal handler does this in daemon mode
    print "Closing DB connection.\n" if $verbose;
    db_disconnect();
  }
}


